"""FuriosaAI optimizer"""

import furiosa.common.utils
from furiosa.optimizer.frontend.onnx import optimize_model

__version__ = furiosa.common.utils.get_sdk_version(__name__)

__all__ = ["optimize_model"]
